# -*- coding: utf-8 -*-

import re
import json
import argparse
from datetime import datetime, timedelta


from ydl_ai_recommender.src.core.manager import OrderDataManager
from ydl_ai_recommender.src.core.manager import ProfileManager
from ydl_ai_recommender.src.core.recommender import UserCFRecommender
from ydl_ai_recommender.src.data.mysql_client import MySQLClient
from ydl_ai_recommender.src.utils import get_conf_path
from ydl_ai_recommender.src.utils.log import create_logger


logger = create_logger(__name__, 'test.log')


def load_local_user_profile():
    """ 为了保证测试结果可重复、同时避免多次执行sql查询操作，用户画像数据使用离线保存的 """
    profile_manager = ProfileManager()
    df = profile_manager.load_test_profile_data()

    user_profile_dict = {}
    for _, row in df.iterrows():
        user_profile_dict[row['uid']] = row

    logger.info('用户画像数 %s ', len(user_profile_dict))
    return user_profile_dict


def load_test_data():
    manager = OrderDataManager()

    # 加载构建召回模型时的用户uid，用于后续判断用户是否是新用户
    train_orders = manager.load_raw_data()
    old_users = set(train_orders['uid'])
    logger.info('订单用户数 %s ', len(old_users))

    # 加载测试数据
    test_orders = manager.load_test_order_data()
    logger.info('加载测试数据成功，共加载 %s 条', len(test_orders))
    return old_users, test_orders


def _sd(n1, n2):
    if n2 == 0:
        return 0
    else:
        return n1 / n2


def evaluation(result_detail):

    top_n_list = [1, 3, 5, 10, 20, 50]
    metrics = {
        'all_test_cnt': 0,
        'all_recall_cnt': 0,
        'old_user_test_cnt': 0,
        'old_user_recall_cnt': 0,
        'new_user_test_cnt': 0,
        'new_user_recall_cnt': 0,
        'same_user_recall_cnt': 0,
        'similar_user_recall_cnt': 0,
        'default_recall_cnt': 0,
        'top_n_recall_cnt': [0] * len(top_n_list),
    }

    for rd in result_detail:
        metrics['all_test_cnt'] += 1
        if rd['is_recall']:
            metrics['all_recall_cnt'] += 1
            
            is_same_user, is_similar_user = False, False
            for top_n, counselor in enumerate(rd['recall_counselors']):
                if counselor['counselor'] == rd['supplier_id']:
                    from_info = counselor['from'].split(' ')
                    if from_info[0] == 'default':
                        metrics['default_recall_cnt'] += 1
                    else:
                        from_id = from_info[1]
                        if from_id == rd['uid']:
                            is_same_user = True
                        if from_id != rd['uid']:
                            is_similar_user = True

                        for i, n in enumerate(top_n_list):
                            if n > top_n:
                                metrics['top_n_recall_cnt'][i] += 1

            if is_same_user:
                metrics['same_user_recall_cnt'] += 1

            if is_similar_user:
                metrics['similar_user_recall_cnt'] += 1

        if rd['is_old_user']:
            metrics['old_user_test_cnt'] += 1
            if rd['is_recall']:
                metrics['old_user_recall_cnt'] += 1
        else:
            metrics['new_user_test_cnt'] += 1
            if rd['is_recall']:
                metrics['new_user_recall_cnt'] += 1

    logger.info('--' * 45)
    logger.info('')
    logger.info('{:<13}{:<7}{:<7}{:<7}'.format('', '样本数', '召回数', '召回率'))
    logger.info('{:<10}{:<10}{:<10}{:<10.2%}'.format('整体\u3000', metrics['all_test_cnt'], metrics['all_recall_cnt'], _sd(metrics['all_recall_cnt'], metrics['all_test_cnt'])))
    logger.info('{:<10}{:<10}{:<10}{:<10.2%}'.format('老用户', metrics['old_user_test_cnt'], metrics['old_user_recall_cnt'], _sd(metrics['old_user_recall_cnt'], metrics['old_user_test_cnt'])))
    logger.info('{:<10}{:<10}{:<10}{:<10.2%}'.format('新用户', metrics['new_user_test_cnt'], metrics['new_user_recall_cnt'], _sd(metrics['new_user_recall_cnt'], metrics['new_user_test_cnt'])))
    
    logger.info('--' * 45)
    logger.info('')
    logger.info('用户自己召回数 {:<4} 占总召回比例 {:.2%}'.format(metrics['same_user_recall_cnt'], _sd(metrics['same_user_recall_cnt'], metrics['all_recall_cnt'])))
    logger.info('相似用户召回数 {:<4} 占总召回比例 {:.2%}'.format(metrics['similar_user_recall_cnt'], _sd(metrics['similar_user_recall_cnt'], metrics['all_recall_cnt'])))
    logger.info('兜底用户召回数 {:<4} 占总召回比例 {:.2%}'.format(metrics['default_recall_cnt'], _sd(metrics['default_recall_cnt'], metrics['all_recall_cnt'])))

    logger.info('--' * 45)
    logger.info('')
    for i, n in enumerate(top_n_list):
        logger.info('top {:<2} 的召回数 {:<4} 召回率 {:.2%}'.format(n, metrics['top_n_recall_cnt'][i], _sd(metrics['top_n_recall_cnt'][i], metrics['all_test_cnt'])))

    return metrics


def do_test(args):
    user_profile_dict = load_local_user_profile()

    try:
        old_users, test_orders = load_test_data()
    except Exception as e:
        logger.error('测试数据加载出错，请确认测试数据已经下载到本地，或启动命令中增加参数 "--do_update_test_data"', exc_info=True)
        return

    recommender = UserCFRecommender(top_n=args.top_n, k=args.k, is_use_db=False)

    result_detail = []
    logger.info('开始测试')
    for index, order_info in test_orders.iterrows():
        if args.max_test > 0:
            if index >= args.max_test:
                break

        uid = order_info['uid']
        profile = user_profile_dict.get(uid)
        if profile is None:
            continue

        is_merge = args.mode == 0
        size = 100 if args.mode == 0 else 0
        recommend_result = recommender.recommend_with_profile(profile, size=size, is_merge=is_merge)
        recall_resons = []
        for rr in recommend_result:
            if rr['counselor'] == order_info['supplier_id']:
                recall_resons.append(rr['from'])

        result_detail.append({
            'uid': uid,
            'supplier_id': order_info['supplier_id'],
            'is_old_user': uid in old_users,
            'recall_counselors': recommend_result,
            'is_recall': len(recall_resons) > 0,
            'recall_reason': '|'.join(recall_resons),
            'update_time': order_info['update_time'],
        })
    logger.info('测试数据推荐完成')

    # 测试结果统计
    logger.info('==' * 20 + ' 测试结果 ' + '==' * 20)
    logger.info('')
    logger.info('相关参数配置: 相似用户数(k) %s ；每个相似用户召回咨询师数(top_n) %s', args.k, args.top_n)

    if args.show_result_by_day:
        days = set(map(lambda x: x['update_time'], result_detail))
        for d in sorted(days):
            logger.info('')
            logger.info('**' * 15 + ' 订单日期 ' + d + '**' * 15)

            evaluation(filter(lambda x: x['update_time'] == d, result_detail))
    else:
        evaluation(result_detail)

    if args.save_test_result:
        # 保存测试结果详情数据
        with open('result_detail.json', 'w', encoding='utf-8') as f:
            json.dump(result_detail, f, ensure_ascii=False, indent=2)


def update_test_data(args):
    """ 更新测试数据，没有就新建 """
    start_date = ''
    if re.match(r'20\d\d-[01]\d-\d\d', args.start_date):
        start_date = args.start_date
    elif re.match(r'-\d+', args.start_date):
        now = datetime.now()
        start_date = (now - timedelta(days=int(args.start_date[1:]))).strftime('%Y-%m-%d')
    elif args.start_date == '0':
        now = datetime.now()
        start_date = now.strftime('%Y-%m-%d')
    else:
        logger.error('args.start_date 参数格式错误，%s', args.start_date)
        raise

    logger.info('开始更新测试数据，测试数据获取条件为 create_time >= %s ', start_date)
    conditions = ['create_time >= "{}"'.format(start_date)]

    client = MySQLClient.create_from_config_file(get_conf_path())
    
    # 订单数据
    manager = OrderDataManager(client)
    manager.update_test_data(conditions=conditions)

    # 用户画像数据
    manager = ProfileManager(client)
    manager.update_test_data(conditions=conditions)
    logger.info('测试数据更新完成')
    

def batch_test(args):
    user_profile_dict = load_local_user_profile()

    try:
        old_users, test_orders = load_test_data()
    except Exception as e:
        logger.error('测试数据加载出错，请确认测试数据已经下载到本地，或启动命令中增加参数 "--do_update_test_data"', exc_info=True)
        return

    def _test(case):
        logger.info('##' * 20 + ' 测试case ' + '##' * 20)
        logger.info(json.dumps(case, ensure_ascii=False))
        logger.info('##' * 20 + ' 测试case ' + '##' * 20)

        recommender = UserCFRecommender(**case)

        result_detail = []
        for index, order_info in test_orders.iterrows():
            if args.max_test > 0:
                if index >= args.max_test:
                    break

            uid = order_info['uid']
            profile = user_profile_dict.get(uid)
            if profile is None:
                continue

            is_merge = args.mode == 0
            size = 100 if args.mode == 0 else 0
            recommend_result = recommender.recommend_with_profile(profile, size=size, is_merge=is_merge)
            recall_resons = []
            for rr in recommend_result:
                if rr['counselor'] == order_info['supplier_id']:
                    recall_resons.append(rr['from'])

            result_detail.append({
                'uid': uid,
                'supplier_id': order_info['supplier_id'],
                'is_old_user': uid in old_users,
                'recall_counselors': recommend_result,
                'is_recall': len(recall_resons) > 0,
                'recall_reason': '|'.join(recall_resons),
                'update_time': order_info['update_time'],
            })
        return result_detail


    def _evaluation(result_detail):
        metrics = evaluation(result_detail)
        metrics['all_recall_ratio'] = _sd(metrics['all_recall_cnt'], metrics['all_test_cnt'])
        metrics['old_user_recall_ratio'] = _sd(metrics['old_user_recall_cnt'], metrics['old_user_test_cnt'])
        metrics['new_user_recall_ratio'] = _sd(metrics['new_user_recall_cnt'], metrics['new_user_test_cnt'])
        metrics['top1'] = _sd(metrics['top_n_recall_cnt'][0], metrics['all_test_cnt'])
        metrics['top10'] = _sd(metrics['top_n_recall_cnt'][3], metrics['all_test_cnt'])
        metrics['top50'] = _sd(metrics['top_n_recall_cnt'][5], metrics['all_test_cnt'])
        return metrics


    test_cases = [
        {'top_n': 2, 'k': 50, 'u2c': 'combination', 'c2c': None, 'is_use_db': False},
        {'top_n': 5, 'k': 20, 'u2c': 'combination', 'c2c': None, 'is_use_db': False},
        {'top_n': 10, 'k': 10, 'u2c': 'combination', 'c2c': None, 'is_use_db': False},
        {'top_n': 20, 'k': 5, 'u2c': 'combination', 'c2c': None, 'is_use_db': False},
        {'top_n': 2, 'k': 50, 'u2c': 'chat', 'c2c': None, 'is_use_db': False},
        {'top_n': 5, 'k': 20, 'u2c': 'chat', 'c2c': None, 'is_use_db': False},
        {'top_n': 10, 'k': 10, 'u2c': 'chat', 'c2c': None, 'is_use_db': False},
        {'top_n': 20, 'k': 5, 'u2c': 'chat', 'c2c': None, 'is_use_db': False},
        {'top_n': 2, 'k': 50, 'u2c': 'order', 'c2c': None, 'is_use_db': False},
        {'top_n': 5, 'k': 20, 'u2c': 'order', 'c2c': None, 'is_use_db': False},
        {'top_n': 10, 'k': 10, 'u2c': 'order', 'c2c': None, 'is_use_db': False},
        {'top_n': 20, 'k': 5, 'u2c': 'order', 'c2c': None, 'is_use_db': False},
        {'top_n': 2, 'k': 50, 'u2c': 'combination', 'c2c': True, 'is_use_db': False},
        {'top_n': 5, 'k': 20, 'u2c': 'combination', 'c2c': True, 'is_use_db': False},
        {'top_n': 10, 'k': 10, 'u2c': 'combination', 'c2c': True, 'is_use_db': False},
        {'top_n': 20, 'k': 5, 'u2c': 'combination', 'c2c': True, 'is_use_db': False},
        {'top_n': 2, 'k': 50, 'u2c': 'chat', 'c2c': True, 'is_use_db': False},
        {'top_n': 5, 'k': 20, 'u2c': 'chat', 'c2c': True, 'is_use_db': False},
        {'top_n': 10, 'k': 10, 'u2c': 'chat', 'c2c': True, 'is_use_db': False},
        {'top_n': 20, 'k': 5, 'u2c': 'chat', 'c2c': True, 'is_use_db': False},
        {'top_n': 2, 'k': 50, 'u2c': 'order', 'c2c': True, 'is_use_db': False},
        {'top_n': 5, 'k': 20, 'u2c': 'order', 'c2c': True, 'is_use_db': False},
        {'top_n': 10, 'k': 10, 'u2c': 'order', 'c2c': True, 'is_use_db': False},
        {'top_n': 20, 'k': 5, 'u2c': 'order', 'c2c': True, 'is_use_db': False},
    ]

    all_test_result = []
    for case in test_cases:
        result_detail = _test(case)
        metrics = _evaluation(result_detail)
        all_test_result.append((case, metrics))

    with open('batch_test_result.tsv', 'w', encoding='utf-8') as f:
        for (case, metrics) in all_test_result:
            items = [case['k'], case['top_n'], case['u2c'], case['c2c']]
            items.extend([metrics['all_recall_ratio'], metrics['old_user_recall_ratio'], metrics['new_user_recall_ratio']])
            items.extend([metrics['top1'], metrics['top10'], metrics['top50']])
            f.write('\t'.join(map(lambda x: str(x), items)) + '\n')


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    
    parser.add_argument('--mode', default=0, type=int, help='模式：0-推荐的咨询师列表去重（默认，与实际线上一样）；1-推荐的咨询师列表没有去重')
    parser.add_argument('--k', default=20, type=int, help='召回相似用户的数量')
    parser.add_argument('--top_n', default=5, type=int, help='每个相似用户召回的咨询师数量')
    parser.add_argument('--max_test', default=0, type=int, help='最多测试数据量')
    parser.add_argument('--save_test_result', default=False, action='store_true', help='保存测试详情结果')
    parser.add_argument('--show_result_by_day', default=False, action='store_true', help='测试结果是否按天展示')
    
    parser.add_argument('--do_batch_test', default=False, action='store_true', help='批量测试')
    
    parser.add_argument('--do_update_test_data', default=False, action='store_true', help='是否更新测试数据')
    parser.add_argument('--start_date', default='-1', type=str, help='测试订单创建的开始时间，可以是"%Y-%m-%d"格式，也可以是 -3 表示前3天')

    args = parser.parse_args()
    logger.info('')

    if args.do_update_test_data:
        logger.info('更新测试数据')
        logger.info('测试数据创建时间 %s', args.start_date)
        update_test_data(args)

    if args.do_batch_test:
        logger.info('执行批量测试')
        batch_test(args)
    else:
        logger.info('开始执行测试任务，测试模式为 %s', args.mode)
        do_test(args)