# -*- coding: utf-8 -*-

import os
import json
from typing import List

import pandas as pd

from ydl_ai_recommender.src.core.profile import encode_user_profile
from ydl_ai_recommender.src.core.counselor_profile import encode_counselor_profile
from ydl_ai_recommender.src.core.manager import DatabaseDataManager
from ydl_ai_recommender.src.utils.log import create_logger


class UserProfileManager(DatabaseDataManager):
    """
    订单用户画像数据管理
    """

    def __init__(self, client=None) -> None:
        super().__init__(client, create_logger(__name__, 'user_profile_manager.log'))
        self.select_items_str = ', '.join([
            'uid', 'country_code', 'channel_id_type', 'ffrom_login', 'user_preference_cate', 'consult_pay_money',
            'listen_pay_money', 'test_items_pay_money', 'course_pay_money', 'consult_order_num',
            'listen_order_num', 'test_items_order_num', 'course_order_num', 'aidi_cst_bias_city',
            'aidi_cst_bias_sex', 'aidi_cst_bias_price', 'aidi_cst_bias_server_type', 'user_login_city',
            'd30_inquire_order_num', 'd30_session_num',
        ])


    def _make_query_sql(self, conditions=None):
        condition_sql = ''
        if conditions:
            condition_sql = ' WHERE ' + ' AND '.join(conditions)

        sql = 'SELECT {} FROM ads.ads_register_user_profiles'.format(self.select_items_str)
        sql += ' WHERE uid IN (SELECT DISTINCT uid FROM ods.ods_ydl_standard_order{})'.format(condition_sql)
        return sql

    def update_data(self):
        """ 从数据库中拉取最新画像特征并保存 """

        sql = self._make_query_sql()
        _, all_data = self.fetch_data_from_db(sql)

        df = pd.DataFrame(all_data)
        self.save_xlsx_data(df, 'all_profile.xlsx')

    def update_test_data(self, conditions):
        """ 从数据库中拉取指定条件画像信息用于测试 """
        
        sql = self._make_query_sql(conditions)
        _, all_data = self.fetch_data_from_db(sql)

        df = pd.DataFrame(all_data)
        self.save_xlsx_data(df, 'test_profile.xlsx')

    def _load_profile_data(self):
        return self.load_xlsx_data('all_profile.xlsx')

    def load_test_profile_data(self):
        return self.load_xlsx_data('test_profile.xlsx')

    def make_embeddings(self):
        user_profiles = self._load_profile_data()
        self.logger.info('订单用户画像数据加载完成，共加载 %s 条', len(user_profiles))
        user_ids, embeddings = [], []
        self.logger.info('开始构建订单用户的用户画像向量')
        for _, profile in user_profiles.iterrows():
            user_ids.append(str(profile['uid']))
            embeddings.append(encode_user_profile(profile))
        
        self.logger.info('用户画像向量构建完成，共构建 %s 用户', len(user_ids))

        with open(os.path.join(self.local_file_dir, 'user_embeddings_ids.txt'), 'w', encoding='utf-8') as f:
            f.write('\n'.join(user_ids))

        with open(os.path.join(self.local_file_dir, 'user_embeddings.json'), 'w', encoding='utf-8') as f:
            json.dump(embeddings, f, ensure_ascii=False)

        return embeddings

    def make_virtual_embedding(self):
        user_ids = []
        embeddings = []

        with open(os.path.join(self.local_file_dir, 'user_embeddings_ids.txt'), 'r', encoding='utf-8') as f:
            user_ids = [line.strip() for line in f]

        with open(os.path.join(self.local_file_dir, 'user_embeddings.json'), 'r', encoding='utf-8') as f:
            embeddings = json.load(f)

        v_embedding_set = {}
        for user_id, embedding in zip(user_ids, embeddings):
            key = '_'.join(map(str, embedding))
            if key not in v_embedding_set:
                v_embedding_set[key] = {
                    'embedding': embedding,
                    'user_ids': [],
                }
            v_embedding_set[key]['user_ids'].append(str(user_id))
        
        v_embedding_list = []
        with open(os.path.join(self.local_file_dir, 'virtual_user_embeddings_ids.txt'), 'w', encoding='utf-8') as f:
            for info in v_embedding_set.values():
                f.write(','.join(info['user_ids']) + '\n')
                v_embedding_list.append(info['embedding'])

        with open(os.path.join(self.local_file_dir, 'virtual_user_embeddings.json'), 'w', encoding='utf-8') as f:
            json.dump(v_embedding_list, f, ensure_ascii=False)


class CounselorProfileManager(DatabaseDataManager):
    """
    咨询师画像数据管理
    """

    def __init__(self, client=None) -> None:
        super().__init__(client, create_logger(__name__, 'counselor_profile_manager.log'))
        self.select_items_str = ', '.join([
            'doctor_id',
            'uid', 'gender', 'age', 'certification', 'education', 'city', 'work_years', 'product_min_price',
            'order_tag_sum', 'order_cate_sum', 'tag_good_at_direction', 'tag_good_at_group', 'tag_good_at_style',
            'tag_good_at_service', 'tag_good_at_language', 'tag_good_at_school', 'cate_good_at',
        ])

    def _make_query_sql(self):
        sql = 'SELECT {} FROM ads.ads_doctor_stats'.format(self.select_items_str)
        sql += ' WHERE is_deleted = 0'
        return sql

    def update_data(self):
        """ 从数据库中拉取最新画像特征并保存 """

        sql = self._make_query_sql()
        _, all_data = self.fetch_data_from_db(sql)

        df = pd.DataFrame(all_data)
        self.save_xlsx_data(df, 'all_counselor_profile.xlsx')

    def _load_profile_data(self):
        return self.load_xlsx_data('all_counselor_profile.xlsx')

    def make_embeddings(self):
        counselor_profiles = self._load_profile_data()
        self.logger.info('咨询师画像数据加载完成，共加载 %s 条', len(counselor_profiles))
        user_ids, embeddings = [], []
        self.logger.info('开始构建咨询师画像向量')
        for _, profile in counselor_profiles.iterrows():
            user_ids.append(str(profile['uid']))
            embeddings.append(encode_counselor_profile(profile))
        
        self.logger.info('咨询师画像向量构建完成，共构建 %s 咨询师', len(user_ids))

        with open(os.path.join(self.local_file_dir, 'counselor_embeddings_ids.txt'), 'w', encoding='utf-8') as f:
            f.write('\n'.join(user_ids))

        with open(os.path.join(self.local_file_dir, 'counselor_embeddings.json'), 'w', encoding='utf-8') as f:
            json.dump(embeddings, f, ensure_ascii=False)

        return embeddings


if __name__ == '__main__':
    # manager = UserProfileManager()
    # manager.make_embeddings()
    # manager.make_virtual_embedding()

    manager = CounselorProfileManager()
    manager.update_data()
    manager.make_embeddings()